"""These are the tiniest elements of the migrations.

An atom consists of both an up and a down Quark.

"""

from __future__ import with_statement

from contextlib import closing


__all__ = ['BaseQuark', 'AdminConnection', 'MigrationConnection', 'AppConnection', 'RawConnection', 'WithCleanConnection']


def Quark(raw_quark):
    """Turns strings and callables into Quarks"""
    if isinstance(raw_quark, basestring):
        return SQLQuark(raw_quark)
    elif isinstance(raw_quark, BaseQuark):
        return raw_quark
    elif callable(raw_quark):
        return PythonQuark(raw_quark)
    else:
        raise ValueError("Unknown atom type")


class BaseQuark(object):
    
    def __init__(self, q):
        self.quark = Quark(q)
        
    def __call__(self, datasource, conn):
        raise Exception("Not implemented")
    
    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self.quark))
    
    __repr__ = __str__


class SQLQuark(BaseQuark):

    def __init__(self, sql):
        self.sql = sql
        
    def __call__(self, datasource, conn):
        return conn.execute(self.sql % datasource.expansions)
    
    def __str__(self):
        return repr(self.sql)

    __repr__ = __str__
    
    
class PythonQuark(BaseQuark):
    
    def __init__(self, func):
        self.func = func
        
    def __call__(self, datasource, conn):
        try:
            return self.func(conn.connection, datasource.expansions)
        except TypeError:
            return self.func(conn.connection)
        
    def __str__(self):
        return self.func.__name__
    
    __repr__ = __str__


class RawConnection(BaseQuark):
    """This is a user level atom that prevents a transaction from occurring."""
    
    def __init__(self, sql):
        if not isinstance(sql, basestring):
            raise TypeError("raw connections must call sql statements")
        self.sql = sql
        
    def __call__(self, datasource, conn):
        with conn.raw_cursor() as c:
            c.execute(self.sql % datasource.expansions)
            
    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.sql))

    __repr__ = __str__


class MigrationConnection(BaseQuark):
    """Runs on the migration connection"""

    def __call__(self, datasource, _conn):
        return self.quark(datasource, datasource.migration_conn)


class AdminConnection(BaseQuark):
    """Runs on the admin connection"""

    def __call__(self, datasource, _conn):
        return self.quark(datasource, datasource.admin_conn)


class AppConnection(BaseQuark):
    """Runs on the app connection"""

    def __call__(self, datasource, _conn):
        return self.quark(datasource, datasource.app_conn)


class WithCleanConnection(BaseQuark):
    """Closes all connections before invoking the enclosed quark"""
    
    def __call__(self, datasource, conn):
        datasource.close()
        return self.quark(datasource, conn)

